# Copyright (c) 2019 SUSE LLC, All rights reserved.
#
# This file is part of obs-img-utils. obs-img-utils provides
# an api and command line utilities for images in OBS.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re

from lxml import html

from urllib.request import (
    urlretrieve,
    urlopen,
    Request
)


class WebContent(object):
    """
    Web Content Scanner and Download interface
    """
    def __init__(self, uri):
        self.uri = uri
        self.namespace_map = dict(
            xhtml='http://www.w3.org/1999/xhtml'
        )

    def fetch_index_list(self, base_name):
        request = Request(self.uri)
        location = urlopen(request)
        tree = html.fromstring(location.read())
        index_list = tree.xpath(
            '//a[starts-with(@href, "{0}")]/@href'.format(base_name),
            namespaces=self.namespace_map
        )
        return sorted(list(set(index_list)))

    def fetch_file(self, base_name, regex, target_file):
        for name in self.fetch_index_list(base_name):
            if re.match(regex, name):
                urlretrieve(
                    os.sep.join([self.uri, name]),
                    target_file
                )
                return name

    def fetch_to_dir(
        self,
        base_name,
        regex,
        target_dir,
        extensions,
        callback=None
    ):
        kwargs = {}
        if callback:
            kwargs['reporthook'] = callback

        for name in self.fetch_index_list(base_name):
            for extension in extensions:
                if name.endswith(extension) and re.match(regex, name):
                    target_file = os.sep.join([target_dir, name])

                    try:
                        urlretrieve(
                            os.sep.join([self.uri, name]),
                            target_file,
                            **kwargs
                        )
                    finally:
                        if callback:
                            callback(0, 0, 0, True)

                    return target_file



# Copyright (c) 2019 SUSE LLC, All rights reserved.
#
# This file is part of obs-img-utils. obs-img-utils provides
# an api and command line utilities for images in OBS.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


class OBSImageUtilsException(Exception):
    """
    Base class to handle all known exceptions.

    Specific exceptions are implemented as sub classes
    of OBSImageUtilsException.

    Attributes
    * :attr:`message`
        Exception message text
    """
    def __init__(self, message):
        self.message = message

    def __str__(self):
        return format(self.message)


class DownloadPackagesFileExceptionOBS(OBSImageUtilsException):
    """
    Exception raised if there is an issue downloading packages file.
    """


class OBSImageConditionsException(OBSImageUtilsException):
    """
    Exception raised if image metadata does not match conditions.
    """


class PackageVersionExceptionOBS(OBSImageUtilsException):
    """
    Exception raised if package does not match version conditions.
    """


class OBSImageChecksumException(OBSImageUtilsException):
    """
    Exception raised if image checksum does not match.
    """


class OBSImageVersionException(OBSImageUtilsException):
    """
    Exception raised when unable to parse image version.
    """


class OBSImageDownloadException(OBSImageUtilsException):
    """
    Exception raised when unable to download image file.
    """



# Copyright (c) 2019 SUSE LLC, All rights reserved.
#
# This file is part of obs-img-utils. obs-img-utils provides
# an api and command line utilities for images in OBS.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import json

from obs_img_utils.cli import main

from click.testing import CliRunner

from unittest.mock import MagicMock, patch

import pytest

urlopen_response = (
    b'<a href="openSUSE-Leap-15.0-Azure.x86_64-1.0.0-Build1.133.packages">'
    b'openSUSE-Leap-15.0-Azure.x86_64-1.0.0-Build1.13..&gt;</a>'
    b'<a href="openSUSE-Leap-15.0-Azure.x86_64-1.0.0-Build1.133.vhdfixed.xz">'
    b'openSUSE-Leap-15.0-Azure.x86_64-1.0.0-Build1.13..&gt;</a>'
    b'<a href="openSUSE-Leap-15.0-Azure.x86_64-1.0.0-Build1.133.'
    b'vhdfixed.xz.sha256">'
    b'openSUSE-Leap-15.0-Azure.x86_64-1.0.0-Build1.13..&gt;</a>'
)


def test_cli_main_help():
    runner = CliRunner()
    result = runner.invoke(main, ['--help'])
    assert result.exit_code == 0
    assert 'The command line interface provides obs ' \
           'image utilities.' in result.output


@pytest.mark.parametrize(
    "endpoint,value",
    [('download',
      'Download image from OBS repository specified by `download-url`'),
     ('packages',
      'Package commands')],
    ids=['obs-img-utils-download', 'obs-img-utils-packages']
)
def test_cli_subcommand_help(endpoint, value):
    runner = CliRunner()
    result = runner.invoke(main, [endpoint, '--help'])
    assert result.exit_code == 0
    assert value in result.output


@patch('obs_img_utils.web_content.urlopen')
@patch('obs_img_utils.web_content.urlretrieve')
@patch('obs_img_utils.api.NamedTemporaryFile')
def test_packages_list(mock_temp_file, mock_retrieve, mock_url_open):
    packages_file = MagicMock()
    packages_file.name = 'tests/data/packages'
    mock_temp_file.return_value = packages_file

    location = MagicMock()
    location.read.return_value = urlopen_response
    mock_url_open.return_value = location

    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            'packages', 'list',
            '--image-name', 'openSUSE-Leap-15.0-Azure',
            '--download-url',
            'https://provo-mirror.opensuse.org/repositories/'
            'Cloud:/Images:/Leap_15.0/images/',
            '--target-dir', 'tests/data'
        ]
    )
    assert result.exit_code == 0

    data = json.loads(result.output)

    assert 'apparmor-parser' in data
    assert data['apparmor-parser']['version'] == '2.12.2'
    assert data['apparmor-parser']['release'] == 'lp150.6.14.1'
    assert data['apparmor-parser']['arch'] == 'x86_64'


@patch('obs_img_utils.web_content.urlopen')
@patch('obs_img_utils.web_content.urlretrieve')
@patch('obs_img_utils.api.NamedTemporaryFile')
def test_packages_show(mock_temp_file, mock_retrieve, mock_url_open):
    packages_file = MagicMock()
    packages_file.name = 'tests/data/packages'
    mock_temp_file.return_value = packages_file

    location = MagicMock()
    location.read.return_value = urlopen_response
    mock_url_open.return_value = location

    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            'packages', 'show',
            '--image-name', 'openSUSE-Leap-15.0-Azure',
            '--download-url',
            'https://provo-mirror.opensuse.org/repositories/'
            'Cloud:/Images:/Leap_15.0/images/',
            '--target-dir', 'tests/data',
            '--package-name', 'apparmor-parser'
        ]
    )
    assert result.exit_code == 0

    data = json.loads(result.output)

    assert data['version'] == '2.12.2'
    assert data['release'] == 'lp150.6.14.1'


@patch('obs_img_utils.api.get_checksum_from_file')
@patch('obs_img_utils.api.get_hash_from_image')
@patch('obs_img_utils.web_content.urlopen')
@patch('obs_img_utils.web_content.urlretrieve')
@patch('obs_img_utils.api.NamedTemporaryFile')
def test_image_download(
    mock_temp_file, mock_retrieve, mock_url_open,
    mock_get_hash_from_image, mock_get_checksum
):
    packages_file = MagicMock()
    packages_file.name = 'tests/data/packages'
    mock_temp_file.return_value = packages_file

    location = MagicMock()
    location.read.return_value = urlopen_response
    mock_url_open.return_value = location

    hash_val = MagicMock()
    hash_val.hexdigest.return_value = 'ABC1234567890'
    mock_get_hash_from_image.return_value = hash_val
    mock_get_checksum.return_value = 'ABC1234567890'

    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            'download',
            '--image-name', 'openSUSE-Leap-15.0-Azure',
            '--download-url',
            'https://provo-mirror.opensuse.org/repositories/'
            'Cloud:/Images:/Leap_15.0/images/',
            '--target-dir', 'tests/data/', '--add-conditions'
        ],
        input='y\n'
              '\n'
              '==\n'
              '1.0.0\n'
              '\n'
              'y\n'
              'apparmor-parser\n'
              '\n'
              '2.12.2\n'
              'lp150.6.14.1\n'
              'n'
    )

    assert result.exit_code == 0
    assert 'Image downloaded: tests/data//openSUSE-Leap-15.0-Azure.x86_64' \
        '-1.0.0-Build1.133.vhdfixed.xz' in result.output


@patch('obs_img_utils.api.time')
@patch('obs_img_utils.api.get_checksum_from_file')
@patch('obs_img_utils.api.get_hash_from_image')
@patch('obs_img_utils.web_content.urlopen')
@patch('obs_img_utils.web_content.urlretrieve')
@patch('obs_img_utils.api.NamedTemporaryFile')
def test_image_download_failed_conditions(
    mock_temp_file, mock_retrieve, mock_url_open,
    mock_get_hash_from_image, mock_get_checksum, mock_time
):
    packages_file = MagicMock()
    packages_file.name = 'tests/data/packages'
    mock_temp_file.return_value = packages_file

    location = MagicMock()
    location.read.return_value = urlopen_response
    mock_url_open.return_value = location

    hash_val = MagicMock()
    hash_val.hexdigest.return_value = 'ABC1234567890'
    mock_get_hash_from_image.return_value = hash_val
    mock_get_checksum.return_value = 'ABC1234567890'

    mock_time.time.side_effect = [0, 0, 1]

    runner = CliRunner()
    result = runner.invoke(
        main,
        [
            'download',
            '--image-name', 'openSUSE-Leap-15.0-Azure',
            '--download-url',
            'https://provo-mirror.opensuse.org/repositories/'
            'Cloud:/Images:/Leap_15.0/images/',
            '--target-dir', 'tests/data/', '--add-conditions',
            '--conditions-wait-time', '1'
        ],
        input='y\n'
              '\n'
              '==\n'
              '1.1.0\n'
              '\n'
              'n\n'
    )

    assert result.exit_code == 1
    assert 'Version condition failed: ' \
           'openSUSE-Leap-15.0-Azure 1.0.0 == 1.1.0' in result.output
    assert 'ImageConditionsException: Image conditions not met' \
        in result.output



# Copyright (c) 2019 SUSE LLC, All rights reserved.
#
# This file is part of obs-img-utils. obs-img-utils provides
# an api and command line utilities for images in OBS.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import click
import logging

from obs_img_utils.utils import (
    get_config,
    click_progress_callback,
    conditions_repl,
    handle_errors,
    echo_package,
    echo_packages,
    get_logger,
    process_shared_options
)
from obs_img_utils.api import OBSImageUtil

shared_options = [
    click.option(
        '-C',
        '--config',
        type=click.Path(exists=True),
        help='OBS Image utils config file to use. Default: '
             '~/.config/obs_img_utils/config.yaml'
    ),
    click.option(
        '--no-color',
        is_flag=True,
        help='Remove ANSI color and styling from output.'
    ),
    click.option(
        '--debug',
        'log_level',
        flag_value=logging.DEBUG,
        help='Display debug level logging to console.'
    ),
    click.option(
        '--verbose',
        'log_level',
        flag_value=logging.INFO,
        default=True,
        help='Display logging info to console. (Default)'
    ),
    click.option(
        '--quiet',
        'log_level',
        flag_value=logging.WARNING,
        help='Disable console output.'
    ),
    click.option(
        '--download-url',
        type=click.STRING,
        help='URL for OBS download repository.'
    ),
    click.option(
        '--target-dir',
        type=click.Path(exists=True),
        help='Directory to store downloaded images and checksums.'
    ),
    click.option(
        '--image-name',
        type=click.STRING,
        help='Image name to download from the download-url.',
        required=True
    ),
    click.option(
        '--arch',
        type=click.Choice(['x86_64', 'aarch64']),
        help='Architecture of the image.'
    ),
    click.option(
        '--profile',
        type=click.STRING,
        help='The multibuild profile name for the image.'
    )
]


def add_options(options):
    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func
    return _add_options


def print_license(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo('GPLv3+')
    ctx.exit()


def abort_if_false(ctx, param, value):
    if not value:
        ctx.abort()


@click.group()
@click.version_option()
@click.option(
    '--license',
    is_flag=True,
    callback=print_license,
    expose_value=False,
    is_eager=True,
    help='Show license information.'
)
@click.pass_context
def main(context):
    """
    The command line interface provides obs image utilities.

    This includes downloading images, checking packages in images
    and getting package version information.
    """
    if context.obj is None:
        context.obj = {}


@click.command()
@click.option(
    '--add-conditions',
    is_flag=True,
    help='Invoke conditions process to specify conditions '
         'for image'
)
@click.option(
    '--conditions-wait-time',
    type=click.INT,
    default=0,
    help='Time (in seconds) to wait for conditions '
         'to be met. Retry period is 150 seconds and'
         ' default is 0 seconds for no wait.'
)
@click.option(
    '--extension',
    type=click.STRING,
    help='Image file extension. Examples: [tar.gz, raw.xz]'
)
@click.option(
    '--checksum-extension',
    type=click.STRING,
    help='Image checksum file extension. Example: sha256'
)
@add_options(shared_options)
@click.pass_context
def download(
    context,
    add_conditions,
    conditions_wait_time,
    extension,
    checksum_extension,
    **kwargs
):
    """
    Download image from OBS repository specified by `download-url`.
    """
    context.obj['conditions_wait_time'] = conditions_wait_time
    context.obj['checksum_extension'] = checksum_extension
    context.obj['extension'] = extension

    process_shared_options(context.obj, kwargs)
    config_data = get_config(context.obj)
    logger = get_logger(config_data.log_level)

    image_conditions = []
    if add_conditions:
        image_conditions = conditions_repl(config_data.no_color)

    with handle_errors(config_data.log_level, config_data.no_color):
        downloader = OBSImageUtil(
            config_data.download_url,
            context.obj['image_name'],
            conditions=image_conditions,
            arch=config_data.arch,
            target_directory=config_data.target_dir,
            profile=config_data.profile,
            log_level=config_data.log_level,
            conditions_wait_time=config_data.conditions_wait_time,
            log_callback=logger,
            report_callback=click_progress_callback,
            checksum_extension=config_data.checksum_extension,
            extension=config_data.extension
        )
        image_source = downloader.get_image()

    click.echo(
        'Image downloaded: {img_source}'.format(img_source=image_source)
    )


@click.group()
def packages():
    """
    Package commands.
    """


@click.command(name='list')
@add_options(shared_options)
@click.pass_context
def list_packages(context, **kwargs):
    """
    Return a list of packages for the given image name.
    """
    process_shared_options(context.obj, kwargs)
    config_data = get_config(context.obj)
    logger = get_logger(config_data.log_level)

    with handle_errors(config_data.log_level, config_data.no_color):
        downloader = OBSImageUtil(
            config_data.download_url,
            config_data.image_name,
            arch=config_data.arch,
            target_directory=config_data.target_dir,
            profile=config_data.profile,
            log_level=config_data.log_level,
            log_callback=logger
        )
        packages_metadata = downloader.get_image_packages_metadata()

    echo_packages(
        packages_metadata,
        no_color=config_data.no_color
    )


@click.command()
@click.option(
    '--package-name',
    type=click.STRING,
    required=True,
    help='Name of the package.'
)
@add_options(shared_options)
@click.pass_context
def show(context, package_name, **kwargs):
    """
    Return information for the provided package name in the given image.
    """
    process_shared_options(context.obj, kwargs)
    config_data = get_config(context.obj)
    logger = get_logger(config_data.log_level)

    with handle_errors(config_data.log_level, config_data.no_color):
        downloader = OBSImageUtil(
            config_data.download_url,
            config_data.image_name,
            arch=config_data.arch,
            target_directory=config_data.target_dir,
            profile=config_data.profile,
            log_level=config_data.log_level,
            log_callback=logger
        )
        packages_metadata = downloader.get_image_packages_metadata()

    echo_package(
        package_name,
        packages_metadata,
        no_color=config_data.no_color
    )


main.add_command(download)
packages.add_command(list_packages)
packages.add_command(show)
main.add_command(packages)



# Copyright (c) 2019 SUSE LLC, All rights reserved.
#
# This file is part of obs-img-utils. obs-img-utils provides
# an api and command line utilities for images in OBS.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pytest

from unittest.mock import patch

from obs_img_utils.api import OBSImageUtil, package_type
from obs_img_utils.exceptions import PackageVersionExceptionOBS


class TestAPI(object):
    def setup(self):
        self.downloader = OBSImageUtil(
            'https://provo-mirror.opensuse.org/repositories/',
            'openSUSE-Leap-15.0-Azure',
            arch='x86_64',
            target_directory='tests/data'
        )

    def test_version_compare(self):
        result = self.downloader._version_compare('1.0', '2.0', '<=')
        assert result

        result = self.downloader._version_compare('1.0', '1.0', '==')
        assert result

        result = self.downloader._version_compare('1.1', '1.0', '>')
        assert result

        result = self.downloader._version_compare('1.0', '2.0', '<')
        assert result

        with pytest.raises(PackageVersionExceptionOBS):
            self.downloader._version_compare('2.0', '1.0', '===')

    @patch.object(OBSImageUtil, '_get_image_checksum')
    @patch('obs_img_utils.api.time')
    def test_wait_for_new_image(self, mock_time, mock_get_checksum):
        self.downloader.image_checksum = 'ABC12345678890'
        mock_get_checksum.return_value = '12345678890ABC'
        self.downloader.wait_for_new_image()

    def test_lookup_package(self):
        packages = {'fake'}
        condition = {'package_name': 'test', 'version': '1.0'}

        # Package does not exist
        result = self.downloader._lookup_package(packages, condition)
        assert result is False

        package = package_type(
            version='0.2',
            release='1.1',
            arch='x86_64',
            checksum='ABC1234567890'
        )
        packages = {'test': package}

        # Version mismatch
        result = self.downloader._lookup_package(packages, condition)
        assert result is False

        condition = {'package_name': 'test', 'release': '1.2'}

        # release mismatch
        result = self.downloader._lookup_package(packages, condition)
        assert result is False
